

#include <wchar.h>
#include "Parser.h"
#include "Scanner.h"


namespace NSQL {


void Parser::SynErr(int n) {
	if (errDist >= minErrDist) errors->SynErr(la->line, la->col, n);
	errDist = 0;
}

void Parser::SemErr(const char* msg) {
	if (errDist >= minErrDist) errors->Error(t->line, t->col, msg);
	errDist = 0;
}

void Parser::Get() {
	for (;;) {
		t = la;
		la = scanner->Scan();
		if (la->kind <= maxT) { ++errDist; break; }

		if (dummyToken != t) {
			dummyToken->kind = t->kind;
			dummyToken->pos = t->pos;
			dummyToken->col = t->col;
			dummyToken->line = t->line;
			dummyToken->next = NULL;
			coco_string_delete(dummyToken->val);
			dummyToken->val = coco_string_create(t->val);
			t = dummyToken;
		}
		la = t;
	}
}

void Parser::Expect(int n) {
	if (la->kind==n) Get(); else { SynErr(n); }
}

void Parser::ExpectWeak(int n, int follow) {
	if (la->kind == n) Get();
	else {
		SynErr(n);
		while (!StartOf(follow)) Get();
	}
}

bool Parser::WeakSeparator(int n, int syFol, int repFol) {
	if (la->kind == n) {Get(); return true;}
	else if (StartOf(repFol)) {return false;}
	else {
		SynErr(n);
		while (!(StartOf(syFol) || StartOf(repFol) || StartOf(0))) {
			Get();
		}
		return StartOf(syFol);
	}
}

void Parser::NSQL() {
		if (la->kind == 7) {
			SelectQuery();
		} else if (la->kind == 10) {
			DeleteQuery();
		} else if (la->kind == 11 || la->kind == 12) {
			QueryPlan();
		} else SynErr(33);
		Expect(6);
}

void Parser::SelectQuery() {
		Expect(7);
		Column();
		while (la->kind == 8) {
			Get();
			Column();
		}
		Expect(9);
		TableSpec();
		if (la->kind == 11 || la->kind == 12) {
			QueryPlan();
		}
}

void Parser::DeleteQuery() {
		Expect(10);
		Expect(9);
		TableSpec();
		OneRowWhereClause();
		dir->flag.allow_delete = 1; 
}

void Parser::QueryPlan() {
		if (la->kind == 11) {
			OneRowWhereClause();
		} else if (la->kind == 12) {
			Scan();
		} else SynErr(34);
}

void Parser::Column() {
		Expect(3);
		char *c_name  = copy_token();          
		char *c_alias = c_name;                
		if (la->kind == 16) {
			Get();
			Expect(3);
			c_alias = copy_token();                
		}
		*dir->visible->new_item() = c_name;    
		*dir->aliases->new_item() = c_alias;   
}

void Parser::TableSpec() {
		if (la->kind == 1) {
			Get();
			dir->database = copy_token(); 
			Expect(17);
		}
		Expect(3);
		dir->table = copy_token(); 
}

void Parser::OneRowWhereClause() {
		Expect(11);
		UniqueIndexSpec();
}

void Parser::Scan() {
		Expect(12);
		idxname = "PRIMARY"; is_bounded = 0;  
		if (la->kind == 13) {
			Get();
			Expect(14);
			dir->flag.table_scan = 1;             
		} else if (la->kind == 23) {
			IndexScan();
			if (la->kind == 15) {
				Get();
				Order();
			}
		} else SynErr(35);
}

void Parser::UniqueIndexSpec() {
		char *idxtype = ""; 
		if (la->kind == 18) {
			Get();
			Expect(19);
			idxtype = "P"; idxname = "*Primary$Key*"; 
		} else if (la->kind == 20) {
			Get();
			Expect(21);
			Expect(3);
			idxtype = "U"; idxname = copy_token();    
		} else SynErr(36);
		build_index(idxtype);                     
		Expect(22);
		ValueList();
}

void Parser::IndexScan() {
		Expect(23);
		Expect(21);
		if (la->kind == 3 || la->kind == 18) {
			if (la->kind == 3) {
				Get();
				idxname = copy_token(); 
			} else {
				Get();
			}
		}
		if (la->kind == 11) {
			WhereClause();
		} else if (la->kind == 6 || la->kind == 15) {
			dir->flag.table_scan = 1;  
			dir->index_scan->name = idxname; 
		} else SynErr(37);
}

void Parser::Order() {
		if (la->kind == 25) {
			Get();
			config::sort_scan(dir,is_bounded,idxname, NSQL::Asc);  
		} else if (la->kind == 26) {
			Get();
			config::sort_scan(dir,is_bounded,idxname, NSQL::Desc); 
		} else SynErr(38);
}

void Parser::ValueList() {
		e.base_col_name = "" ; e.rel_op = 4; 
		IndexValue();
		while (la->kind == 8) {
			Get();
			IndexValue();
		}
}

void Parser::IndexValue() {
		if (la->kind == 3 || la->kind == 31) {
			if (la->kind == 31) {
				Get();
			}
			Expect(3);
			e.vtype = NSQL::Param; e.value = copy_token(); 
		} else if (la->kind == 4 || la->kind == 5) {
			constant();
			e.vtype = NSQL::Const; e.value = copy_token(); 
		} else SynErr(39);
		index_condition(); 
}

void Parser::WhereClause() {
		Expect(11);
		build_index("O");  
		is_bounded = 1;    
		IndexCondition();
		while (la->kind == 24) {
			Get();
			IndexCondition();
		}
}

void Parser::IndexCondition() {
		Expect(3);
		e.base_col_name = copy_token(); 
		relop();
		IndexValue();
}

void Parser::relop() {
		if (la->kind == 22) {
			Get();
			e.rel_op = 4;  
		} else if (la->kind == 27) {
			Get();
			e.rel_op = 3;  
		} else if (la->kind == 28) {
			Get();
			e.rel_op = 2;  
		} else if (la->kind == 29) {
			Get();
			e.rel_op = 1;  
		} else if (la->kind == 30) {
			Get();
			e.rel_op = 0;  
		} else SynErr(40);
}

void Parser::constant() {
		if (la->kind == 4) {
			Get();
		} else if (la->kind == 5) {
			Get();
		} else SynErr(41);
}



void Parser::Parse() {
	t = NULL;
	la = dummyToken = new Token();
	la->val = coco_string_create(L"Dummy Token");
	Get();
	NSQL();

}

Parser::Parser(Scanner *scanner) {
	_EOF = 0;
	_DBName = 1;
	_FuncName = 2;
	_Name = 3;
	_qstring = 4;
	_number = 5;
	maxT = 32;

	minErrDist = 2;
	errDist = minErrDist;
	this->scanner = scanner;
	errors = new Errors();
}

bool Parser::StartOf(int s) {
	const bool T = true;
	const bool x = false;

	static bool set[1][34] = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x}
	};



	return set[s][la->kind];
}

Parser::~Parser() {
	delete errors;
	delete dummyToken;
}

Errors::Errors() {
	count = 0;
}

void Errors::SynErr(int line, int col, int n) {
	wchar_t* s;
	switch (n) {
			case 0: s = coco_string_create(L"EOF expected"); break;
			case 1: s = coco_string_create(L"DBName expected"); break;
			case 2: s = coco_string_create(L"FuncName expected"); break;
			case 3: s = coco_string_create(L"Name expected"); break;
			case 4: s = coco_string_create(L"qstring expected"); break;
			case 5: s = coco_string_create(L"number expected"); break;
			case 6: s = coco_string_create(L"\";\" expected"); break;
			case 7: s = coco_string_create(L"\"select\" expected"); break;
			case 8: s = coco_string_create(L"\",\" expected"); break;
			case 9: s = coco_string_create(L"\"from\" expected"); break;
			case 10: s = coco_string_create(L"\"delete\" expected"); break;
			case 11: s = coco_string_create(L"\"where\" expected"); break;
			case 12: s = coco_string_create(L"\"using\" expected"); break;
			case 13: s = coco_string_create(L"\"table\" expected"); break;
			case 14: s = coco_string_create(L"\"scan\" expected"); break;
			case 15: s = coco_string_create(L"\"order\" expected"); break;
			case 16: s = coco_string_create(L"\"as\" expected"); break;
			case 17: s = coco_string_create(L"\".\" expected"); break;
			case 18: s = coco_string_create(L"\"primary\" expected"); break;
			case 19: s = coco_string_create(L"\"key\" expected"); break;
			case 20: s = coco_string_create(L"\"unique\" expected"); break;
			case 21: s = coco_string_create(L"\"index\" expected"); break;
			case 22: s = coco_string_create(L"\"=\" expected"); break;
			case 23: s = coco_string_create(L"\"ordered\" expected"); break;
			case 24: s = coco_string_create(L"\"and\" expected"); break;
			case 25: s = coco_string_create(L"\"asc\" expected"); break;
			case 26: s = coco_string_create(L"\"desc\" expected"); break;
			case 27: s = coco_string_create(L"\"<\" expected"); break;
			case 28: s = coco_string_create(L"\"<=\" expected"); break;
			case 29: s = coco_string_create(L"\">\" expected"); break;
			case 30: s = coco_string_create(L"\">=\" expected"); break;
			case 31: s = coco_string_create(L"\"$\" expected"); break;
			case 32: s = coco_string_create(L"??? expected"); break;
			case 33: s = coco_string_create(L"invalid NSQL"); break;
			case 34: s = coco_string_create(L"invalid QueryPlan"); break;
			case 35: s = coco_string_create(L"invalid Scan"); break;
			case 36: s = coco_string_create(L"invalid UniqueIndexSpec"); break;
			case 37: s = coco_string_create(L"invalid IndexScan"); break;
			case 38: s = coco_string_create(L"invalid Order"); break;
			case 39: s = coco_string_create(L"invalid IndexValue"); break;
			case 40: s = coco_string_create(L"invalid relop"); break;
			case 41: s = coco_string_create(L"invalid constant"); break;

		default:
		{
			wchar_t format[20];
			coco_swprintf(format, 20, L"error %d", n);
			s = coco_string_create(format);
		}
		break;
	}

        char *cstring = coco_string_create_char(s);
        log_err(http_server,"[Col %d] %s", col, cstring);
	coco_string_delete(s);
        delete cstring;
	count++;
}

void Errors::Error(int line, int col, const char *s) {
        log_err(http_server,"[Col %d] %s", col, s);
	count++;
}

void Errors::Warning(int line, int col, const wchar_t *s) {
	wprintf(L"-- line %d col %d: %ls\n", line, col, s);
}

void Errors::Warning(const wchar_t *s) {
	wprintf(L"%ls\n", s);
}

void Errors::Exception(const wchar_t* s) {
	wprintf(L"%ls", s); 
	exit(1);
}

}; // namespace

