

#if !defined(COCO_PARSER_H__)
#define COCO_PARSER_H__

#include "mod_ndb.h"


#include "Scanner.h"

namespace NSQL {


class Errors {
public:
	int count;			// number of errors detected

	Errors();
	void SynErr(int line, int col, int n);
	void Error(int line, int col, const char *s);
	void Warning(int line, int col, const wchar_t *s);
	void Warning(const wchar_t *s);
	void Exception(const wchar_t *s);
        server_rec *http_server;

}; // Errors

class Parser {
private:
	int _EOF;
	int _DBName;
	int _FuncName;
	int _Name;
	int _qstring;
	int _number;
	int maxT;

	Token *dummyToken;
	int errDist;
	int minErrDist;

	void SynErr(int n);
	void Get();
	void Expect(int n);
	bool StartOf(int s);
	void ExpectWeak(int n, int follow);
	bool WeakSeparator(int n, int syFol, int repFol);

public:
	Scanner *scanner;
	Errors  *errors;

	Token *t;			// last recognized token
	Token *la;			// lookahead token

cmd_parms *cmd;
config::dir *dir;
NSQL::Expr e;
char *idxname;
const char *cf_err;
int is_bounded;

char * copy_token() { return coco_create_apache_string(cmd->pool, t->val); }

void build_index(char *idxtype) {
  if(config::get_index_by_name(dir, idxname) == -1)
    config::build_index_record(cmd, dir, idxtype, idxname);
}

void index_condition()  {
  const char *cf_err = 0;
  if(e.vtype == NSQL::Param) 
    cf_err = config::named_idx(cmd, dir, idxname, &e);
  else if(e.vtype == NSQL::Const)
    cf_err = config::index_constant(cmd, dir, idxname, &e);
  else assert(0);
  if(cf_err) SemErr(cf_err);
}  
  


	Parser(Scanner *scanner);
	~Parser();
	void SemErr(const char *msg);

	void NSQL();
	void SelectQuery();
	void DeleteQuery();
	void QueryPlan();
	void Column();
	void TableSpec();
	void OneRowWhereClause();
	void Scan();
	void UniqueIndexSpec();
	void IndexScan();
	void Order();
	void ValueList();
	void IndexValue();
	void WhereClause();
	void IndexCondition();
	void relop();
	void constant();

	void Parse();

}; // end Parser

}; // namespace

#endif // !defined(COCO_PARSER_H__)

